@@ -1,5 +1,5 @@
 /*
- *	$Id: proc.c,v 1.2 1999/04/18 19:08:12 mj Exp $
+ *	$Id: proc.c,v 1.7 2000/01/20 21:14:44 mj Exp $
  *
  *	The PCI Library -- Configuration Access via /proc/bus/pci
  *
@@ -19,40 +19,63 @@
 
 #include "internal.h"
 
-#include <asm/unistd.h>
-#if defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ < 1
-#include <syscall-list.h>
-#endif
-
 /*
- * As libc doesn't support pread/pwrite yet, we have to call them directly
- * or use lseek/read/write instead.
+ *  We'd like to use pread/pwrite for configuration space accesses, but
+ *  unfortunately it isn't simple at all since all libc's until glibc 2.1
+ *  don't define it.
  */
-#if !(defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ > 0)
 
-#if defined(__GLIBC__) && !(defined(__powerpc__) && __GLIBC__ == 2 && __GLIBC_MINOR__ == 0)
+#if defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ > 0
+/* glibc 2.1 or newer -> pread/pwrite supported automatically */
+
+#elif defined(i386) && defined(__GLIBC__)
+/* glibc 2.0 on i386 -> call syscalls directly */
+#include <asm/unistd.h>
+#include <syscall-list.h>
 #ifndef SYS_pread
-#define SYS_pread __NR_pread
+#define SYS_pread 180
 #endif
-static int
-pread(unsigned int fd, void *buf, size_t size, loff_t where)
-{
-  return syscall(SYS_pread, fd, buf, size, where);
-}
-
+static int pread(unsigned int fd, void *buf, size_t size, loff_t where)
+{ return syscall(SYS_pread, fd, buf, size, where); }
 #ifndef SYS_pwrite
-#define SYS_pwrite __NR_pwrite
+#define SYS_pwrite 181
 #endif
-static int
-pwrite(unsigned int fd, void *buf, size_t size, loff_t where)
-{
-  return syscall(SYS_pwrite, fd, buf, size, where);
-}
+static int pwrite(unsigned int fd, void *buf, size_t size, loff_t where)
+{ return syscall(SYS_pwrite, fd, buf, size, where); }
+
+#elif defined(i386)
+/* old libc on i386 -> call syscalls directly the old way */
+#include <asm/unistd.h>
+static _syscall5(int, pread, unsigned int, fd, void *, buf, size_t, size, u32, where_lo, u32, where_hi);
+static _syscall5(int, pwrite, unsigned int, fd, void *, buf, size_t, size, u32, where_lo, u32, where_hi);
+static int do_read(struct pci_dev *d __attribute__((unused)), int fd, void *buf, size_t size, int where) { return pread(fd, buf, size, where, 0); }
+static int do_write(struct pci_dev *d __attribute__((unused)), int fd, void *buf, size_t size, int where) { return pwrite(fd, buf, size, where, 0); }
+#define HAVE_DO_READ
+
 #else
-static _syscall4(int, pread, unsigned int, fd, void *, buf, size_t, size, loff_t, where);
-static _syscall4(int, pwrite, unsigned int, fd, void *, buf, size_t, size, loff_t, where);
+/* In all other cases we use lseek/read/write instead to be safe */
+#define make_rw_glue(op) \
+	static int do_##op(struct pci_dev *d, int fd, void *buf, size_t size, int where)	\
+	{											\
+	  struct pci_access *a = d->access;							\
+	  int r;										\
+	  if (a->fd_pos != where && lseek(fd, where, SEEK_SET) < 0)				\
+	    return -1;										\
+	  r = op(fd, buf, size);								\
+	  if (r < 0)										\
+	    a->fd_pos = -1;									\
+	  else											\
+	    a->fd_pos = where + r;								\
+	  return r;										\
+	}
+make_rw_glue(read)
+make_rw_glue(write)
+#define HAVE_DO_READ
 #endif
 
+#ifndef HAVE_DO_READ
+#define do_read(d,f,b,l,p) pread(f,b,l,p)
+#define do_write(d,f,b,l,p) pwrite(f,b,l,p)
 #endif
 
 static void
@@ -95,7 +118,7 @@ static void
 proc_scan(struct pci_access *a)
 {
   FILE *f;
-  char buf[256];
+  char buf[512];
 
   if (snprintf(buf, sizeof(buf), "%s/devices", a->method_params[PCI_ACCESS_PROC_BUS_PCI]) == sizeof(buf))
     a->error("File name too long");
@@ -105,13 +128,13 @@ proc_scan(struct pci_access *a)
   while (fgets(buf, sizeof(buf)-1, f))
     {
       struct pci_dev *d = pci_alloc_dev(a);
-      unsigned int dfn, vend;
+      unsigned int dfn, vend, cnt, known;
 
-      sscanf(buf,
-#ifdef HAVE_64BIT_ADDRESS
-	     "%x %x %x %llx %llx %llx %llx %llx %llx %llx",
+      cnt = sscanf(buf,
+#ifdef HAVE_LONG_ADDRESS
+	     "%x %x %x %llx %llx %llx %llx %llx %llx %llx %llx %llx %llx %llx %llx %llx %llx",
 #else
-	     "%x %x %x %lx %lx %lx %lx %lx %lx %lx",
+	     "%x %x %x %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx",
 #endif
 	     &dfn,
 	     &vend,
@@ -122,14 +145,31 @@ proc_scan(struct pci_access *a)
 	     &d->base_addr[3],
 	     &d->base_addr[4],
 	     &d->base_addr[5],
-	     &d->rom_base_addr);
+	     &d->rom_base_addr,
+	     &d->size[0],
+	     &d->size[1],
+	     &d->size[2],
+	     &d->size[3],
+	     &d->size[4],
+	     &d->size[5],
+	     &d->rom_size);
+      if (cnt != 9 && cnt != 10 && cnt != 17)
+	a->error("proc: parse error (read only %d items)", cnt);
       d->bus = dfn >> 8U;
       d->dev = PCI_SLOT(dfn & 0xff);
       d->func = PCI_FUNC(dfn & 0xff);
       d->vendor_id = vend >> 16U;
       d->device_id = vend & 0xffff;
-      d->known_fields = a->buscentric ? PCI_FILL_IDENT
-				      : (PCI_FILL_IDENT | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE);
+      known = PCI_FILL_IDENT;
+      if (!a->buscentric)
+	{
+	  known |= PCI_FILL_IRQ | PCI_FILL_BASES;
+	  if (cnt >= 10)
+	    known |= PCI_FILL_ROM_BASE;
+	  if (cnt >= 17)
+	    known |= PCI_FILL_SIZES;
+	}
+      d->known_fields = known;
       pci_link_dev(a, d);
     }
   fclose(f);
@@ -153,6 +193,7 @@ proc_setup(struct pci_dev *d, int rw)
       if (a->fd < 0)
 	a->warning("Cannot open %s", buf);
       a->cached_dev = d;
+      a->fd_pos = 0;
     }
   return a->fd;
 }
@@ -165,7 +206,7 @@ proc_read(struct pci_dev *d, int pos, byte *buf, int len)
 
   if (fd < 0)
     return 0;
-  res = pread(fd, buf, len, pos);
+  res = do_read(d, fd, buf, len, pos);
   if (res < 0)
     {
       d->access->warning("proc_read: read failed: %s", strerror(errno));
@@ -187,7 +228,7 @@ proc_write(struct pci_dev *d, int pos, byte *buf, int len)
 
   if (fd < 0)
     return 0;
-  res = pwrite(fd, buf, len, pos);
+  res = do_write(d, fd, buf, len, pos);
   if (res < 0)
     {
       d->access->warning("proc_write: write failed: %s", strerror(errno));
